#!/usr/bin/python
#
# Copyright 2015 Christian Forssen
#
#  This file is part of WIGXJPF.
#
#  WIGXJPF is free software: you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  WIGXJPF is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with WIGXJPF.  If not, see
#  <http://www.gnu.org/licenses/>.
#

import pywigxjpf as wig

import sys

failure = int(sys.argv[1])

print('WIGXJPF python error handling test (%d)' % failure)
sys.stdout.flush()

# Initialize
if (failure != 1):
    wig.wig_table_init(2*100 if failure != 2 else 2*1,
                       9     if failure != 3 else -1)
if (failure != 4):
    wig.wig_temp_init(2*100 if failure != 5 else 1)

# Note that arguments are in two_j = 2*j.
val3j = wig.wig3jj(2* 10 , 2* 15 , 2* 10 ,\
                   2*(-3), 2* 12 , 2*(-9))
if failure == 6:
    val3j = wig.wig3jj(2* 1000 , 2* 1500 , 2* 1000 ,\
                       2*(-300), 2* 1200 , 2*(-900))
print('3J(10  15  10; -3  12  -9): %.12g' % val3j)
sys.stdout.flush()

val6j = wig.wig6jj(2* 10 , 2* 15 , 2* 10 ,\
                   2*  7,  2*  7 , 2*  9)
if failure == 7:
    val6j = wig.wig6jj(2* 1000 , 2* 1500 , 2* 1000 ,\
                       2*  700,  2*  700 , 2*  900)
print('6J{10  15  10;  7   7   9}: %.12g' % val6j)
sys.stdout.flush()

val9j = wig.wig9jj( 1,  2,  3,\
                    4,  6,  8, \
                    3,  6,  9 )
if failure == 8:
    val9j = wig.wig9jj( 100,  200,  300,\
                        400,  600,  800, \
                        300,  600,  900 )
print('9J{0.5 1 1.5; 2 3 4; 1.5 3 4.5}: %.12g' % val9j)
sys.stdout.flush()

print('Using the (numpy) array interface')
sys.stdout.flush()

val3j = wig.wig3jj_array([2* 10 , 2* 15 , 2* 10 ,\
                          2*(-3), 2* 12 , 2*(-9)])
if failure == 9:
    val3j = wig.wig3jj_array([2* 1000 , 2* 1500 , 2* 1000 ,\
                              2*(-300), 2* 1200 , 2*(-900)])
print('3J(10  15  10; -3  12  -9): %.12g' % val3j)
sys.stdout.flush()

val6j = wig.wig6jj_array([2* 10 , 2* 15 , 2* 10 ,\
                          2*  7,  2*  7 , 2*  9])
if failure == 10:
    val6j = wig.wig6jj_array([2* 1000 , 2* 1500 , 2* 1000 ,\
                              2*  700,  2*  700 , 2*  900])
print('6J{10  15  10;  7   7   9}: %.12g' % val6j)
sys.stdout.flush()

val9j = wig.wig9jj_array([1,  2,  3,\
                          4,  6,  8, \
                          3,  6,  9])
if failure == 11:
    val9j = wig.wig9jj_array([100,  200,  300,\
                              400,  600,  800, \
                              300,  600,  900])
print('9J{0.5 1 1.5; 2 3 4; 1.5 3 4.5}: %.12g' % val9j)
sys.stdout.flush()

# Free memory space
wig.wig_temp_free()
wig.wig_table_free()




#!/usr/bin/python
#
# Copyright 2015 Christian Forssen
#
#  This file is part of WIGXJPF.
#
#  WIGXJPF is free software: you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  WIGXJPF is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with WIGXJPF.  If not, see
#  <http://www.gnu.org/licenses/>.
#

import pywigxjpf as wig

print('WIGXJPF python test program')

# Initialize
wig.wig_table_init(2*100,9)
wig.wig_temp_init(2*100)

# Note that arguments are in two_j = 2*j.
val3j = wig.wig3jj(2* 10 , 2* 15 , 2* 10 ,\
                   2*(-3), 2* 12 , 2*(-9))
print('3J(10  15  10; -3  12  -9): %.12g' % val3j)

val6j = wig.wig6jj(2* 10 , 2* 15 , 2* 10 ,\
                   2*  7,  2*  7 , 2*  9)
print('6J{10  15  10;  7   7   9}: %.12g' % val6j)

val9j = wig.wig9jj( 1,  2,  3,\
                    4,  6,  8, \
                    3,  6,  9 )
print('9J{0.5 1 1.5; 2 3 4; 1.5 3 4.5}: %.12g' % val9j)

print('Using the (numpy) array interface')

val3j = wig.wig3jj_array([2* 10 , 2* 15 , 2* 10 ,\
                          2*(-3), 2* 12 , 2*(-9)])
print('3J(10  15  10; -3  12  -9): %.12g' % val3j)

val6j = wig.wig6jj_array([2* 10 , 2* 15 , 2* 10 ,\
                          2*  7,  2*  7 , 2*  9])
print('6J{10  15  10;  7   7   9}: %.12g' % val6j)

val9j = wig.wig9jj_array([1,  2,  3,\
                          4,  6,  8, \
                          3,  6,  9])
print('9J{0.5 1 1.5; 2 3 4; 1.5 3 4.5}: %.12g' % val9j)

# Free memory space
wig.wig_temp_free()
wig.wig_table_free()




from cffi import FFI
ffibuilder = FFI()

# When compiled by 'python pywigxjpf_ffi_builder.py' according to
# Makefile (README), we are in this directory, so need a ../ prefix.
# When compiling from setup.py, the working directory is the parent.

updir = ""
if __name__ == "__main__": # "__cffi__" when compiling from setup
    updir = "../"

# ffibuilder.set_source("pywigxjpf_ffi",
#    r""" /* Compiled code.  All is in the library. */
#         /* The header is included to check the shared prototypes. */
#         #include "wigxjpf.h"
#     """,
#    include_dirs=[updir + 'inc/'],
#    extra_objects=[updir + 'lib/libwigxjpf.a'])

ffibuilder.set_source("pywigxjpf_ffi",
    r""" /* Compiled code. */
         /* The header is included to check the shared prototypes. */
         #include "wigxjpf.h"
     """,
    include_dirs=[updir + x for x in
                  # 'Normal' (user) compile will use the gen/ version
                  # of wigxjpf_auto_config.h.  PyPI compiles will
                  # include from pywigxjpf/pregen/
                  ['inc/', 'src/', 'cfg/', 'gen/', 'pywigxjpf/pregen/']],
    sources = [updir + 'src/' + x for x in
               ['c_wrap.c',
                # 'c_wrap_float128.c', # float128 is not exposed
                'calc.c',
                # 'calc_float128.c', # float128 is not exposed
                # 'fortran_wrap.c', # not needed for python interface
                'prime_factor.c',
                'trivial_zero.c',
                'wigxjpf_error.c']
    ],
    # Perform different error handling on e.g. missing temp arrays. */
    define_macros=[('PYWIGXJPF_ERROR_HANDLING', '1')],
    # libraries=['quadmath'] # float128 is not exposed
)

ffibuilder.cdef("""
    /* Declarations shared from C to Python. */
    double wig3jj(int two_j1, int two_j2, int two_j3,
                  int two_m1, int two_m2, int two_m3);
    double wig6jj(int two_j1, int two_j2, int two_j3,
                  int two_m1, int two_m2, int two_m3);
    double wig9jj(int two_j1, int two_j2, int two_j3,
                  int two_j4, int two_j5, int two_j6,
                  int two_j7, int two_j8, int two_j9);
    void wig_table_init(int max_two_j, int wigner_type);
    void wig_table_free(void);
    void wig_temp_init(int max_two_j);
    void wig_temp_free(void);
""")

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



try:
    from pywigxjpf import \
        wig_table_init, wig_temp_init, wig_temp_free, wig_table_free, \
        wig3jj, wig6jj, wig9jj, wig3jj_array, wig6jj_array, wig9jj_array
except ImportError:
    from pywigxjpf.pywigxjpf import \
        wig_table_init, wig_temp_init, wig_temp_free, wig_table_free, \
        wig3jj, wig6jj, wig9jj, wig3jj_array, wig6jj_array, wig9jj_array



#
# Copyright 2015 Christian Forssen
#
#  This file is part of WIGXJPF.
#
#  WIGXJPF is free software: you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  WIGXJPF is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with WIGXJPF.  If not, see
#  <http://www.gnu.org/licenses/>.
#

"""
Filename: pywigxjpf.py
Python interface to the wigxjpf (dynamic) library with the use of cffi.
Defines seven functions:

    (1) wig_table_init(max_two_j,wigner_type)
    (2) wig_table_free()
    (3) wig_temp_init(max_two_j)
    (4) wig_temp_free()
    (5) wig3jj(jj1,jj2,jj3,
               mm1,mm2,mm3)
    (6) wig6jj(jj1,jj2,jj3,
               jj4,jj5,jj6)
    (7) wig9jj(jj1,jj2,jj3,
               jj4,jj5,jj6,
               jj7,jj8,jj9)

Note that the arguments are to be given as integers, with
twice the numeric value (this is what jj tries to indicate).
I.e. half-integer arguments will be passed as odd integers.

The two init functions must be called before evaluating any
symbol.

In addition, interfaces that take an array with the arguments
are also provided.  They match the previously provided
interface, but are renamed:

    (8)  wig3jj_array([jj1,jj2,jj3,
                       mm1,mm2,mm3])
    (9)  wig6jj_array([jj1,jj2,jj3,
                       jj4,jj5,jj6])
    (10) wig9jj_array([jj1,jj2,jj3,
                       jj4,jj5,jj6,
                       jj7,jj8,jj9])
"""

__version__ = '1.13.1  '
__author__ = 'C. Forssen and H.T. Johansson'
__all__ = ['pywigxjpf']

import numpy as np

try:
    from pywigxjpf_ffi import ffi, lib
except ImportError:
    from pywigxjpf.pywigxjpf_ffi import ffi, lib

wig_table_init = lib.wig_table_init
wig_table_free = lib.wig_table_free

wig_temp_init  = lib.wig_temp_init
wig_temp_free  = lib.wig_temp_free

wig3jj = lib.wig3jj
wig6jj = lib.wig6jj
wig9jj = lib.wig9jj

def wig3jj_array(b):
    jj1,jj2,jj3,mm1,mm2,mm3 = np.asarray(b,dtype=np.intc)
    return lib.wig3jj(jj1,jj2,jj3,mm1,mm2,mm3)

def wig6jj_array(b):
    jj1,jj2,jj3,jj4,jj5,jj6 = np.asarray(b,dtype=np.intc)
    return lib.wig6jj(jj1,jj2,jj3,jj4,jj5,jj6)

def wig9jj_array(b):
    jj1,jj2,jj3,jj4,jj5,jj6,jj7,jj8,jj9 = np.asarray(b,dtype=np.intc)
    return lib.wig9jj(jj1,jj2,jj3,jj4,jj5,jj6,jj7,jj8,jj9)

if __name__ == '__main__':
    print('WIGXJPF python test facility')

    # Initialize
    wig_table_init(2*100, 9)
    wig_temp_init(2*100)

    # Note that arguments are in two_j = 2*j.

    # First use the CFFI style direct calls:
    val3j = wig3jj(2* 10 , 2* 15 , 2* 10 ,\
                   2*(-3), 2* 12 , 2*(-9))
    print('3J(10  15  10; -3  12  -9): %.12g' % val3j)

    val6j = wig6jj(2* 10 , 2* 15 , 2* 10 ,\
                   2*  7,  2*  7 , 2*  9 )
    print('6J{10  15  10;  7   7   9}: %.12g' % val6j)

    val9j = wig9jj( 1,  2,  3, \
                    4,  6,  8, \
                    3,  6,  9 )
    print('9J{0.5 1 1.5; 2 3 4; 1.5 3 4.5}: %.12g' % val9j)

    # Then the wrapped calls, taking an array as argument:
    val3j = wig3jj_array([2* 10 , 2* 15 , 2* 10 ,\
                          2*(-3), 2* 12 , 2*(-9)])
    print('3J(10  15  10; -3  12  -9): %.12g' % val3j)

    val6j = wig6jj_array([2* 10 , 2* 15 , 2* 10 ,\
                          2*  7,  2*  7 , 2*  9 ])
    print('6J{10  15  10;  7   7   9}: %.12g' % val6j)

    val9j = wig9jj_array( [1,  2,  3, \
                           4,  6,  8, \
                           3,  6,  9] )
    print('9J{0.5 1 1.5; 2 3 4; 1.5 3 4.5}: %.12g' % val9j)

    # Free memory space
    wig_temp_free()
    wig_table_free()



#!/usr/bin/env python

from setuptools import setup, find_packages

try:
    import pycparser as dummy
except ImportError:
    print('##############################################\n'
          '!        ERROR! Caught import error.         !\n'
          '!      Please, install pycparser first:      !\n'
          '!                                            !\n'
          '!            pip install pycparser           !\n'
          '!                                            !\n'
          '##############################################\n\n'
          'For details, see '
          'https://github.com/pypa/setuptools/issues/391#issuecomment-202919511'
          '\n\n')

setup(name='pywigxjpf',
      version='1.13.1  ',
      description='Wigner Symbols',
      long_description="""
WIGXJPF evaluates Wigner 3j, 6j and 9j symbols
accurately using prime factorisation and multi-word integer arithmetic.

Known issue
-----------

Please install pycparser first.

(See https://github.com/pypa/setuptools/issues/391#issuecomment-202919511 )

Inline usage information
------------------------

Available inline::

    import pywigxjpf as wig
    help(wig)                # For interfaces.
    help(wig.pywigxjpf)      # For usage information.

Library usage
-------------

The python interface to wigxjpf uses cffi.

Defines seven functions::

    wig_table_init(max_two_j,wigner_type)
    wig_table_free()
    wig_temp_init(max_two_j)
    wig_temp_free()

    wig3jj(jj1,jj2,jj3, mm1,mm2,mm3)
    wig6jj(jj1,jj2,jj3, jj4,jj5,jj6)
    wig9jj(jj1,jj2,jj3, jj4,jj5,jj6, jj7,jj8,jj9)

Note that the arguments are to be given as integers, with
twice the numeric value (this is what jj tries to indicate).
I.e. half-integer arguments will be passed as odd integers.

The two init functions must be called before evaluating any
symbol.

In addition, interfaces that take an array with the arguments
are also provided::

    wig3jj_array([jj1,jj2,jj3, mm1,mm2,mm3])
    wig6jj_array([jj1,jj2,jj3, jj4,jj5,jj6])
    wig9jj_array([jj1,jj2,jj3, jj4,jj5,jj6, jj7,jj8,jj9])
""",
      long_description_content_type='text/x-rst',
      license = "GPLv3",
      author='C. Forssen and H. T. Johansson',
      url='http://fy.chalmers.se/subatom/wigxjpf',
      packages=['pywigxjpf'],
      # package_data={'pywigxjpf': ['../lib/libwigxjpf_shared.*']},
      setup_requires=['numpy',"cffi>=1.0.0"],
      cffi_modules=["pywigxjpf/pywigxjpf_ffi_builder.py:ffibuilder"],
      install_requires=['numpy',"cffi>=1.0.0"],
     )



from pywigxjpf_ffi import ffi, lib
import pywigxjpf_ffi
import time

import numba
#from numba import cffi_support  # Line below works on python2 also.
from numba.core.typing import cffi_utils as cffi_support
cffi_support.register_module(pywigxjpf_ffi)
nb_wig3jj = pywigxjpf_ffi.lib.wig3jj

lib.wig_table_init(100,9);
lib.wig_temp_init(100);

print(lib.wig3jj(5,6,7,3,4,-7))
print(lib.wig6jj(6,6,6,6,6,6))
print(lib.wig9jj(6,6,6,7,7,6,7,7,8))
print(lib.wig9jj(6,6,6,7,7,6,7,7,8))
print(lib.wig6jj(6,6,6,6,6,int(4+0.5*4)))

# Benchmark 1, make calls for a lot of trivially-0 symbols.
@numba.jit(nopython=True)
def benchmark(jjmax):
    sum = 0.0
    calls = 0
    for jj1 in range(0, jjmax + 1):
        for jj2 in range(0, jjmax + 1):
            for jj3 in range(0, jjmax + 1):
                for mm1 in range(-jjmax, jjmax + 1):
                    for mm2 in range(-jjmax, jjmax + 1):
                        for mm3 in range(-jjmax, jjmax + 1):
                            w = nb_wig3jj(jj1, jj2, jj3, \
                                           mm1, mm2, mm3)
                            # print((w)
                            sum = sum + w
                            calls = calls+1
    return (sum,calls)

# Benchmark 2, avoiding trivial 0 by triangle rules.
@numba.jit(nopython=True)
def benchmark_opt(jjmax):
    sum = 0.0
    calls = 0
    for jj1 in range(0, jjmax + 1, 1):
        for jj2 in range(0, jjmax + 1, 1):
            jj3_min = abs(jj1-jj2)
            jj3_max = jj1+jj2
            if (jj3_max > jjmax):
                jj3_max = jjmax
            for jj3 in range(jj3_min, jj3_max + 1, 2):
                for mm1 in range(-jj1, jj1 + 1, 2):
                    for mm2 in range(-jj2, jj2 + 1, 2):
                        #for m3 in range(-j3, j3 + 1):
                        mm3 = -mm1-mm2
                        if (abs(mm3) <= jjmax):
                            w = nb_wig3jj(jj1, jj2, jj3, \
                                           mm1, mm2, mm3)
                            sum = sum + w
                            calls = calls+1
    return (sum,calls)

jjmax=10

for i in range(0,5):
    start_time = time.time()
    wigsum, total_calls = benchmark(jjmax)
    total_time = time.time()-start_time
    print("Benchmark 1 for jjmax=%d, sum=%.10f, time=%.5fs, "
          "time/call=%4.0fns [%d calls]" %
          (jjmax, wigsum, total_time, total_time/total_calls*1e9, total_calls))

for i in range(0,5):
    start_time = time.time()
    wigsum, total_calls = benchmark_opt(jjmax)
    total_time = time.time()-start_time
    print("Benchmark 2 for jjmax=%d, sum=%.10f, time=%.5fs, "
          "time/call=%4.0fns [%d calls]" %
          (jjmax, wigsum, total_time, total_time/total_calls*1e9, total_calls))

lib.wig_temp_free();
lib.wig_table_free();

print('Done')

# i7-8550U:   (with jjmax=20)

#                  incl. trivial-0    excl. trivial-0
#
#  Total:          638277381 calls        357599 calls
#
#  cffi   numba         16 ns/call        273 ns/call
#  C      native        10 ns/call        263 ns/call

# E3-1240 v3:  (with jjmax=10)

#                  incl. trivial-0    excl. trivial-0
#
#  Total:           12326391 calls        15475 calls
#
#  cffi   direct       188 ns/call        486 ns/call
#  cffi   numba         16 ns/call        185 ns/call
#  C      native        10 ns/call        175 ns/call
